'''
C-Language Backend Code for Simulating Room Impulse Response

author: github.com/ludlows
2019-11 

This program is designed with the hope that it will be useful, but WITHOUT ANY GUARANTEE.
'''
# expose the rir function
from .cyrir import rir
